/* $Id: ide-std.c,v 1.4 1999/06/11 14:29:45 ralf Exp $
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * IDE routines for typical pc-like standard configurations.
 *
 * Copyright 2008 TiVo Inc. All Rights Reserved. by Ralf Baechle
 */
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ide.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>
#include <asm/ptrace.h>
#include <asm/hdreg.h>
#include <asm/io.h>

#include <linux/bootmem.h>
#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <asm/reboot.h>
#include <asm/traps.h>
#include <asm/page.h>
#include <asm/delay.h>
#include <asm/bcm7038.h>
#include <linux/tivoconfig.h>

static void (*tivo_cmd_hook)(ide_drive_t *, unsigned char *);
static unsigned int tivo_size_priv(int dma_chunks, int block_chunks);
static void tivo_init_drvpriv(struct idetivo_request *trq, int reading);
static int tivo_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading);
static void tivo_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static void tivo_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static void tivo_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)); 

static struct ll_dma_ops tivo_sgops = {
    size_priv   :   tivo_size_priv,
    init_drvpriv:   tivo_init_drvpriv,
    add_dmachunk:   tivo_add_dmachunk,
    add_blkchunk:   tivo_add_blkchunk,
    end_blkchunk:   tivo_end_blkchunk,
    set_cmd_hook:   tivo_set_cmd_hook,
};

struct tivo_sgent {
    u32 address;
    u32 count;
};

struct desc_chunk {
    unsigned int sglist;   //bus address of scatter list
};

struct tivo_ide_priv {
    struct desc_chunk *first_chunk;
    struct desc_chunk *cur_chunk;
    struct tivo_sgent *first_sgent;
    struct tivo_sgent *cur_sgent;
    int big_flush;
    int last_size;
    u32 bus_addr_end;
};

struct sataWatch {
    unsigned long status;
    unsigned long error;
    unsigned long history;
    unsigned long mask;
    unsigned long control;
};

static struct sataWatchBoth {
    struct sataWatch iSata;
    struct sataWatch eSata;
} sataWatchMemorized;

/*
 These functions are used to preprocess TiVo I/O requests

 Report back how much room needed to allocate along with the request
 for use by the DMA engine, based on number of scatter list entries,
 and number of distinct scatter lists.  Each separate IDE transfer
 (of max MAX_NRSECTORS sectors) requires a separate scatter list.
*/
static unsigned int tivo_size_priv(int dma_chunks, int block_chunks)
{
   unsigned int total_chunk_size = dma_chunks * sizeof(struct tivo_sgent);

   unsigned int size = sizeof(struct tivo_ide_priv) +
               block_chunks * sizeof(struct desc_chunk) +
               (block_chunks - 1) * 8 + // worst case PRD table alignment
               L1_CACHE_BYTES - 4 +  // worst case cache line alignment
               total_chunk_size;
    return size;
}

/*
  put the chunk list right after the priv struct
  put the scatter list after the chunk list (but cache line align it)
*/
static void tivo_init_drvpriv(struct idetivo_request *trq, int reading)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    unsigned long tmp;

    priv->first_chunk = (struct desc_chunk *)((unsigned long)priv +
                                              sizeof(struct tivo_ide_priv));

    tmp = (unsigned long)(priv->first_chunk + trq->num_chunks);
    priv->first_sgent = (struct tivo_sgent *)L1_CACHE_ALIGN(tmp);

    priv->cur_sgent = priv->first_sgent;
    priv->cur_chunk = priv->first_chunk;
    priv->cur_chunk->sglist = virt_to_phys(priv->cur_sgent);

    // this is a bit of platform specificness, but there's no generic way
    // to do it: if the request totals more than twice the dcache size,
    // mark to do one big flush, instead of flushing each individual chunk
    //
    priv->big_flush = (trq->req->nr_sectors > 16 * 1024 / SECTOR_BYTES);

    priv->bus_addr_end = 1; // an impossible addr
}


/*
    Add a chunk to the DMA scatter list and do any necessary cache flushing
    Both kernel and user virtual addresses are passed in, in case the user
    address is needed for flushing purposes (to avoid virtual aliasing
    problem on CPUs with virually indexed caches).  The user address is
    guaranteed not to page fault and size is guaranteed to be non-zero.
*/
static int tivo_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    struct tivo_sgent *sgent = priv->cur_sgent;
    u32 bus_addr;
    int total_size;

    if ((k_addr | size) & 3) {
        printk(KERN_WARNING "tivo_ide: DMA not word aligned\n");
        return -EINVAL;
    }

    if (!priv->big_flush) {
        if (reading) {
            dma_cache_inv(u_addr, size);
        } else {
            dma_cache_wback_inv(u_addr, size);
        }
    }

    bus_addr = (u32)virt_to_bus((void *)k_addr);
    // Consolidate adjacent chunks
    //
    // Note that total size will never exceed 128K, because we start a
    // new scatter list at least every 128K
    //
    total_size = size;
    u32 cur_len = total_size;


    while (cur_len) {
        u32 xcount, bcount = 0x10000 - (bus_addr & 0xffff);
        
        if (bcount > cur_len)
            bcount = cur_len;

        xcount = bcount & 0xffff;

#ifdef CONFIG_TIVO_CPU_TO_BE32
        sgent->address = cpu_to_be32(bus_addr);
        sgent->count = cpu_to_be32(xcount);
#else
        sgent->address = cpu_to_le32(bus_addr);
        sgent->count = cpu_to_le32(xcount);
#endif /* CONFIG_TIVO_CPU_TO_BE32 */
        bus_addr += bcount;
        cur_len -= bcount;

        priv->cur_sgent = sgent + 1;
        sgent = priv->cur_sgent;
    }
    priv->last_size = total_size;
    priv->bus_addr_end = bus_addr + size;
    return 0;
}
// Finish up a block of DMA chunks that makes up a single IDE transfer,
// and start a new one
//
static void tivo_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    struct desc_chunk *desc = priv->cur_chunk++;

#ifdef CONFIG_TIVO_CPU_TO_BE32
    (priv->cur_sgent - 1)->count |= cpu_to_be32(0x80000000);
#else
    (priv->cur_sgent - 1)->count |= cpu_to_le32(0x80000000);
#endif
    // Align start of PRD table to 16-bytes
    priv->cur_sgent = (struct tivo_sgent *)
        (((unsigned long)priv->cur_sgent + 0xf) & ~0xf);
    (desc + 1)->sglist = virt_to_phys(priv->cur_sgent);

    priv->bus_addr_end = 1; // an impossible addr
}

// Finish up the last block of DMA chunks, and do any final prep on the
// data going to the DMA engine
//
static void tivo_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    unsigned long sg_start = (unsigned long)priv->first_sgent;
    unsigned long sg_end = (unsigned long)priv->cur_sgent;

#ifdef CONFIG_TIVO_CPU_TO_BE32
    (priv->cur_sgent - 1)->count |= cpu_to_be32(0x80000000);
#else
    (priv->cur_sgent - 1)->count |= cpu_to_le32(0x80000000);
#endif
    if (!priv->big_flush) {
        dma_cache_wback_inv(sg_start, sg_end - sg_start);
    } else {
        flush_dcache();
    }
    priv->cur_chunk = priv->first_chunk;
#ifdef CONFIG_TIVO_IDE_ARC
    if (trq->req->cmd == READ) ARC_start(sg_start, sg_end-sg_start);
#endif
}

static void tivo_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)) {
    tivo_cmd_hook = handler;
}

ide_startstop_t bcm2_dma_intr(ide_drive_t *drive)
{
    int i, dma_stat;
    byte stat;
    struct request *rq = HWGROUP(drive)->rq;
    struct idetivo_request *trq = rq->tivo;

#ifdef CONFIG_TIVO_IDE_ARC
    ARC_end();
#endif
    dma_stat = HWIF(drive)->dmaproc(ide_dma_end, drive);
    stat = GET_STAT();            /* get drive status */
    if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
        if (!dma_stat) {
            plog_diskstop(rq);
            if (trq) {
                unsigned int max_nrsectors = trq->max_nrsectors;

                if (rq->nr_sectors > max_nrsectors) {
                    ide_hwif_t *hwif = HWIF(drive);
                    ide_startstop_t startstop;

                    rq->nr_sectors -= max_nrsectors;
                    rq->sector += max_nrsectors;
                    trq->progress += max_nrsectors * SECTOR_BYTES;
                    add_blkdev_randomness(MAJOR(rq->rq_dev));
                    SELECT_DRIVE(hwif, drive);
                    if (ide_wait_stat(&startstop, drive, drive->ready_stat,
                                      BUSY_STAT|DRQ_STAT, WAIT_READY)) {
                        printk("%s: drive not ready for command\n", drive->name);
                        return startstop;
                    }
                    return DRIVER(drive)->do_request(drive, rq, rq->sector
                                                      + trq->partition_start);
                } else {
                    trq->progress += rq->nr_sectors * SECTOR_BYTES;
                ide_end_request(1, HWGROUP(drive));
                trq->status = IDETIVO_DONE;
                return ide_stopped;
                }
            } else {
                for (i = rq->nr_sectors; i > 0;) {
                    i -= rq->current_nr_sectors;
                    ide_end_request(1, HWGROUP(drive));
                }
            }
            return ide_stopped;
        }
        printk("%s: dma_intr: bad DMA status\n", drive->name);
    }
    if (trq) {
        struct tivo_ide_priv *priv = trq->drv_priv;

        // if the request gets retried, we need to reprocess this
        // chunk of the DMA scatterlist
        priv->cur_chunk--;
        trq->status = IDETIVO_ERROR;
        trq->drive_status = stat;
        // XXX: should probably read drive regs to query how far we
        // got in last 256 sector (max) chunk, too...
    }
    return ide_error(drive, "dma_intr", stat);
}

int bcm2_build_dmatable(ide_drive_t *drive) 
{
    struct request *rq = HWGROUP(drive)->rq;
    
    if (!rq->tivo) {
        return 1;
    }
    struct tivo_ide_priv *priv = rq->tivo->drv_priv;
    struct desc_chunk *chunk = priv->cur_chunk++;
    ide_hwif_t *hwif = HWIF(drive);

    rq->tivo->status = IDETIVO_INPROGRESS;

    unsigned long dma_base        = hwif->dma_base;
    /* write the PRD table to the controller */
    outl(cpu_to_le32((unsigned long)chunk->sglist), dma_base + 4); 
    barrier();

    return 1;
}

#ifdef BCM_IDE_DEBUG
static inline void print_sg_list(struct tivo_ide_priv *priv)
{
    struct tivo_sgent *sgent = priv->first_sgent;
    unsigned long sg_end = (unsigned long)priv->cur_sgent;

    while ((unsigned long)sgent != sg_end) {
        unsigned long sg_address = be32_to_cpu(sgent->address);
        unsigned long sg_count = sgent->count;
        if (sg_count & 0x80000000) {
            sg_count &= 0x7FFFFFFF;
            sg_count = be32_to_cpu(sg_count);
        }
        printk("sg_address = 0x%x sg_count = %d\n", sg_address, sg_count);
        sgent++;
    }
}
#endif /* BCM_IDE_DEBUG */

static int tivo_ide_default_irq(ide_ioreg_t base)
{
	return 0;
}

static ide_ioreg_t tivo_ide_default_io_base(int index)
{
	return 0;
}

static void tivo_ide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port,
                                    ide_ioreg_t ctrl_port, int *irq)
{
	ide_ioreg_t reg = data_port;
	int i;
	TivoConfigValue ide_cap;

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg;
		reg += 1;
	}
	if (ctrl_port) {
		hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	} else {
		hw->io_ports[IDE_CONTROL_OFFSET] = hw->io_ports[IDE_DATA_OFFSET] + 0x206;
	}
	if (irq != NULL)
		*irq = 0;
	hw->io_ports[IDE_IRQ_OFFSET] = 0;
#if 1	
	if (GetTivoConfig(kTivoConfigIdeCap, &ide_cap)) {
	    ide_hwifs[0].udma_four = ide_hwifs[1].udma_four = 0;
	} else {
	    ide_cap &= kTivoConfigIdeCapAta66;
	    ide_hwifs[0].udma_four = ide_hwifs[1].udma_four = ide_cap ? 1 : 0;
	}
#endif
	ide_hwifs[0].tivo.version = 
            ide_hwifs[1].tivo.version = IDETIVO_DRIVER_VERSION;
	ide_hwifs[0].tivo.sgops = 
            ide_hwifs[1].tivo.sgops = &tivo_sgops;
}

static int tivo_ide_request_irq(unsigned int irq,
                                void (*handler)(int,void *, struct pt_regs *),
                                unsigned long flags, const char *device,
                                void *dev_id)
{
	return request_irq(irq, handler, flags, device, dev_id);
}			

static void tivo_ide_free_irq(unsigned int irq, void *dev_id)
{
	free_irq(irq, dev_id);
}

static int tivo_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
	return check_region(from, extent);
}

static void tivo_ide_request_region(ide_ioreg_t from, unsigned int extent,
                                    const char *name)
{
	request_region(from, extent, name);
}

static void tivo_ide_release_region(ide_ioreg_t from, unsigned int extent)
{
	release_region(from, extent);
}

struct ide_ops bcm_ide_ops = {
	&tivo_ide_default_irq,
	&tivo_ide_default_io_base,
	&tivo_ide_init_hwif_ports,
	&tivo_ide_request_irq,
	&tivo_ide_free_irq,
	&tivo_ide_check_region,
	&tivo_ide_request_region,
	&tivo_ide_release_region
};

void _tivo_ide_out_byte(u8 data, ide_ioreg_t port)
{
  *((volatile unsigned char *)(mips_io_port_base + port)) = __ioswab32(data);
}

void _tivo_ide_out_word(u16 data, ide_ioreg_t port)
{
  /* 
   * Uh, two spaces per tab?, then hard tabs in bcm_ide_ops?, 
   * What random standard are we writing code to now? 
   * Pick a freaking standard for the file, and stick to it.
   * Learn contextual coding and stop giving folks headaches.
   * I don't know who did it, and don't want to know, when you
   * read this..., fix it, and remove this comment.
   *
   * --snay
   */
  *((volatile unsigned short *)(mips_io_port_base + port)) = __ioswab32(data);
}

u8 _tivo_ide_in_byte(ide_ioreg_t port)
{
  return  __ioswab32(*((volatile unsigned char *)(mips_io_port_base + port)));
}

u16 _tivo_ide_in_word(ide_ioreg_t port)
{
  return  __ioswab32(*((volatile unsigned short *)(mips_io_port_base + port)));
}
     

/*
 * These are duplicates of code elsewhere, we need it here because it's
 * #ifndef CONFIG_TIVO'ed where TiVo platforms need it to be different,
 * but for PX, we actually need the standard PCI IDE code:
 */

#undef  insw
#define insw(port, addr, count) ide_insw(port, addr, count)
static inline void ide_insw(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(u16 *)addr =  __ioswab32(*(volatile u16 *)(mips_io_port_base + port));
    addr += 2;
  }
}

#undef  outw
#define outw(port, addr, count) ide_outw(port, addr, count)
static inline void ide_outw(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(volatile u16 *)(mips_io_port_base + port) =  __ioswab32(*(u16 *) addr);
    addr += 2;
  }
}

#undef  insl
#define insl(port, addr, count) ide_insl(port, addr, count)
static inline void ide_insl(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(u32 *)addr =  __ioswab32(*(volatile u32 *)(mips_io_port_base + port));
    addr += 4;
  }
}

#undef  outl
#define outl(port, addr, count) ide_outl(port, addr, count)
static inline void ide_outl(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(volatile u32 *)(mips_io_port_base + port) =  __ioswab32(*(u32 *) addr);
    addr += 4;
  }
}


/*
 * This is used for most PIO data transfers *from* the IDE interface
 */
void ide_input_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	byte io_32bit = drive->io_32bit;

	if (io_32bit) {
		insl(IDE_DATA_REG, buffer, wcount);
	} else {
		insw(IDE_DATA_REG, buffer, wcount<<1);
	}
}

/*
 * This is used for most PIO data transfers *to* the IDE interface
 */
void ide_output_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	byte io_32bit = drive->io_32bit;

	if (io_32bit) {
		outsl(IDE_DATA_REG, buffer, wcount);
	} else {
		outsw(IDE_DATA_REG, buffer, wcount<<1);
	}
}

/*
 * The following routines are mainly used by the ATAPI drivers.
 *
 * These routines will round up any request for an odd number of bytes,
 * so if an odd bytecount is specified, be sure that there's at least one
 * extra byte allocated for the buffer.
 */
void atapi_input_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
	++bytecount;
	ide_input_data (drive, buffer, bytecount / 4);
	if ((bytecount & 0x03) >= 2)
		insw (IDE_DATA_REG, ((byte *)buffer) + (bytecount & ~0x03), 1);
}

void atapi_output_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
	++bytecount;
	ide_output_data (drive, buffer, bytecount / 4);
	if ((bytecount & 0x03) >= 2)
		outsw (IDE_DATA_REG, ((byte *)buffer) + (bytecount & ~0x03), 1);
}

/*
 * *****************************************************************
 * Stuff to check dynamic plug and un-plug of SATA drive.
 * *****************************************************************/

#include <asm/bcm7038.h>

void mmio_outl(unsigned long port, unsigned int val)
{
	/* Need to swap because the output is in little endiam mode
	 * even though PCI controller is configured in big-endian.
	 */
	*__preg32( ((unsigned int)SATA_MEM_BASE) + (port)) = swab32(val);
}

unsigned int mmio_inl(unsigned long port)
{
	/* Need to swap because the output is in little endiam mode
	 * even though PCI controller is configured in big-endian.
	 */
	return swab32(*__preg32(((unsigned int) SATA_MEM_BASE)+ port));
}

EXPORT_SYMBOL(mmio_outl);
EXPORT_SYMBOL(mmio_inl);

static DECLARE_WAIT_QUEUE_HEAD(sata_wait_int);

static int mmio_sata_error_offsets[NUM_SATA_PORTS] =
	{
		MMIO_SATA_ERROR,
		MMIO_SATA_ERROR + PORT_OFFSET
	};

static void GetStatusOfSATA (struct sataWatchBoth *status) {
    status->iSata.status = mmio_inl(MMIO_SATA_STATUS);
    status->iSata.error = mmio_inl(MMIO_SATA_ERROR);
    status->iSata.mask = mmio_inl(MMIO_SATA_MASK);
    status->iSata.control = mmio_inl(MMIO_SATA_CONTROL);
    status->eSata.status = mmio_inl(PORT_OFFSET + MMIO_SATA_STATUS);
    status->eSata.error = mmio_inl(PORT_OFFSET + MMIO_SATA_ERROR);
    status->eSata.mask = mmio_inl(PORT_OFFSET + MMIO_SATA_MASK);
    status->eSata.control = mmio_inl(PORT_OFFSET + MMIO_SATA_CONTROL);
}

static void PrintStatusOfSATA (struct sataWatchBoth *status) {
    printk("iSATA: status %08lX error %08lX history %08lX mask %08lX control %08lX\n",
           status->iSata.status, status->iSata.error, status->iSata.history, status->iSata.mask, status->iSata.control);
    printk("eSATA: status %08lX error %08lX history %08lX mask %08lX control %08lX\n",
           status->eSata.status, status->eSata.error, status->eSata.history, status->eSata.mask, status->eSata.control);
}


int		sata_link_level_error_count = 0;
extern int sata_link_level_reset_count;

/*
 * This function resets the entire SATA dricves and also the entire link 
 */ 
void SATALinkLevelReset(void) {
	volatile u32 	sata_ctrl;
	const u32		bit18mask = 0x1 << 18;
	
	if (sata_link_level_reset_count == 0) {
		sata_ctrl = *(volatile u32*)(BCM_SUN_TOP_CTRL_SATA_CTRL);
		sata_ctrl |= bit18mask;
		*(volatile u32*)(BCM_SUN_TOP_CTRL_SATA_CTRL) = sata_ctrl;
		mdelay(2);
		*(volatile u32*)(BCM_SUN_TOP_CTRL_SATA_CTRL) = sata_ctrl ^ bit18mask;
		mdelay(2);
		sata_link_level_error_count = 0;
		sata_link_level_reset_count = 0;
	}
}

int  SATAErrorChangeTimeout(void) {
	if (sata_link_level_error_count > 5) {
		return 1;
	}
	return 0;
}

void SATALinkErrorThresholdCheck(void) {
	if (sata_link_level_error_count > 5) {
		SATALinkLevelReset();
	}
}


void CheckStatusOfSATA (ide_hwgroup_t *hwgroup) {
    struct sataWatchBoth newStatus;
    unsigned long bing;
    newStatus.iSata.history = sataWatchMemorized.iSata.history;
    newStatus.eSata.history = sataWatchMemorized.eSata.history;
    GetStatusOfSATA (&newStatus);
    if (sataWatchMemorized.iSata.status != newStatus.iSata.status ||
        sataWatchMemorized.iSata.error != newStatus.iSata.error ||
        sataWatchMemorized.iSata.mask != newStatus.iSata.mask ||
        sataWatchMemorized.iSata.control != newStatus.iSata.control ||
        sataWatchMemorized.eSata.status != newStatus.eSata.status ||
        sataWatchMemorized.eSata.error != newStatus.eSata.error ||
        sataWatchMemorized.eSata.mask != newStatus.eSata.mask ||
        sataWatchMemorized.eSata.control != newStatus.eSata.control) {
        sataWatchMemorized = newStatus;
        PrintStatusOfSATA(&newStatus);
        /*
         * Acknowledge/clear error bits other than PHY-status-change, so that we can see
         * whether any of these errors is being triggered repeatedly
         */
        if (0 != (bing = sataWatchMemorized.iSata.error & MMIO_PHY_WATCH_MASK)) {
            sataWatchMemorized.iSata.history |= bing;
            mmio_outl(MMIO_SATA_ERROR, bing);
			++sata_link_level_error_count;
        }
        if (0 != (bing = sataWatchMemorized.eSata.error & MMIO_PHY_WATCH_MASK)) {
            sataWatchMemorized.eSata.history |= bing;
            mmio_outl(MMIO_SATA_ERROR + PORT_OFFSET, bing);
			++sata_link_level_error_count;
        }
		if (sata_link_level_error_count > 5) {
			if (hwgroup != NULL)
				ide_reset_command_timer(hwgroup);
		}
    }
}

void InitializeStatusOfSATA() {
    GetStatusOfSATA (&sataWatchMemorized);
    PrintStatusOfSATA(&sataWatchMemorized);
    /*
     * Clear all of the status bits during startup, so that we can see any changes
     * which may occur after this point.  Doing so will trigger a change report the
     * first time we call CheckStatusOfSATA.
     */
    mmio_outl(MMIO_SATA_ERROR, MMIO_PHY_RDY_CHANGE | MMIO_PHY_WATCH_MASK);
    mmio_outl(MMIO_SATA_ERROR + PORT_OFFSET, MMIO_PHY_RDY_CHANGE | MMIO_PHY_WATCH_MASK);
    sataWatchMemorized.iSata.history = sataWatchMemorized.eSata.history = 0;
    sata_link_level_error_count = 0;
	sata_link_level_reset_count = 0;
}

void sata_error_interrupt_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int regvalue, i;

        CheckStatusOfSATA((ide_hwgroup_t *)dev_id);

	for(i=0; i< NUM_SATA_PORTS; i++)
	{
		regvalue = mmio_inl(mmio_sata_error_offsets[i]);

		/* Acknowledge the interrupt */
		regvalue &= MMIO_PHY_RDY_CHANGE;
		mmio_outl(mmio_sata_error_offsets[i], regvalue);
		if (regvalue)
		{
			/* If it is plug/un-plug interrupt, wakeup any
			 * process that is waiting on this event.
			 */
			//printk("Waking up the sata_wait ****\n");
			wake_up_interruptible(&sata_wait_int);
		}
	}

	return 0;
}

unsigned int sata_wait_for_IRQ(int port, int offset, int flag)
{
	port = port*PORT_OFFSET;

	/* Need to clear the bit first and then wait on interrupt */
	mmio_outl(MMIO_SATA_ERROR+port, MMIO_PHY_RDY_CHANGE);
	mmio_outl(MMIO_SATA_MASK+port, MMIO_PHY_RDY_CHANGE);

	/* We dont have to clear the interrupt as it is taken care by 
	 * interrupt handler.
	 */
	return wait_event_interruptible(sata_wait_int, ((mmio_inl(offset)) & flag));
}

EXPORT_SYMBOL(sata_error_interrupt_handler);
EXPORT_SYMBOL(sata_wait_for_IRQ);

#ifdef CONFIG_TIVO_IDE_ARC

#define BVNF_INTR2_2_R5F_CLEAR              *(volatile unsigned int *)(0xb0103208)
#define BVNF_INTR2_2_R5F_MASK_STATUS        *(volatile unsigned int *)(0xb010320c)
#define BVNF_INTR2_2_R5F_MASK_SET           *(volatile unsigned int *)(0xb0103210)
#define BVNF_INTR2_2_R5F_MASK_CLEAR         *(volatile unsigned int *)(0xb0103214)

#define MEMC_ARC_0_CNTRL                    *(volatile unsigned int *)(0xb0106400)
#define MEMC_ARC_0_ADRS_RANGE_LOW           *(volatile unsigned int *)(0xb0106404)
#define MEMC_ARC_0_ADRS_RANGE_HIGH          *(volatile unsigned int *)(0xb0106408)
#define MEMC_ARC_0_READ_RIGHTS_LOW          *(volatile unsigned int *)(0xb010640c)
#define MEMC_ARC_0_READ_RIGHTS_HIGH         *(volatile unsigned int *)(0xb0106410)
#define MEMC_ARC_0_WRITE_RIGHTS_LOW         *(volatile unsigned int *)(0xb0106414)
#define MEMC_ARC_0_WRITE_RIGHTS_HIGH        *(volatile unsigned int *)(0xb0106418)
#define MEMC_ARC_0_VIOLATION_INFO_LOW       *(volatile unsigned int *)(0xb010641c)
#define MEMC_ARC_0_VIOLATION_INFO_HIGH      *(volatile unsigned int *)(0xb0106420)

#define HIF_RAC 41          // Client for 'Host Interface Read Ahead Cache', aka the cpu

extern void show_registers(struct pt_regs *regs);

static void arc_isr(int irq, void *dev_id, struct pt_regs *regs)
{
    BVNF_INTR2_2_R5F_MASK_SET=~0;   // mask it

    printk("IDE ARC violation at %.8X in %.8X-%.8X, EPC=%lX\n",
            MEMC_ARC_0_VIOLATION_INFO_LOW*8,
            MEMC_ARC_0_ADRS_RANGE_LOW*8, 
            MEMC_ARC_0_ADRS_RANGE_HIGH*8, 
            regs->cp0_epc);
    
    // show_registers(regs);
}

void ARC_init(void)
{
    printk("Initializing IDE ARC\n"); 
    MEMC_ARC_0_CNTRL=0;              // disable ARC
    BVNF_INTR2_2_R5F_MASK_SET=~0;   // mask it
    MEMC_ARC_0_READ_RIGHTS_LOW=(HIF_RAC<32)?~(1<<HIF_RAC):~0;
    MEMC_ARC_0_READ_RIGHTS_HIGH=(HIF_RAC<32)?~0:~(1<<(HIF_RAC-32));
    MEMC_ARC_0_WRITE_RIGHTS_LOW=(HIF_RAC<32)?~(1<<HIF_RAC):~0;
    MEMC_ARC_0_WRITE_RIGHTS_HIGH=(HIF_RAC<32)?~0:~(1<<(HIF_RAC-32));
     
    if (request_irq(BCM_IRQ_BVNF_CPU_INTR_2, arc_isr, SA_INTERRUPT, "IDE ARC", NULL))
        printk(KERN_ERR "WARNING: can't register IDE ARC interrupt, proceeding without it...\n");
    else
        MEMC_ARC_0_CNTRL=0x0a;      // enable read and write check
}

void ARC_start(int addr, int len)
{
    if (MEMC_ARC_0_CNTRL && (BVNF_INTR2_2_R5F_MASK_STATUS & 1))   // if enabled and masked...
    {
        if ((addr & 0x40000007) || ((addr+len) & 0x40000000) || len < 8)   
        {
            printk(KERN_WARNING "IDE ARC unusable %.8X %X\n", addr, len);
            return;
        }
        
        int physaddr = addr & 0x1fffffff;       // ensure physical
        MEMC_ARC_0_ADRS_RANGE_LOW=physaddr/8;
        MEMC_ARC_0_ADRS_RANGE_HIGH=(physaddr+len-1)/8;
        BVNF_INTR2_2_R5F_CLEAR=~0;              // kill any pending 
        (void)MEMC_ARC_0_VIOLATION_INFO_HIGH;   // unlatch violation
        BVNF_INTR2_2_R5F_MASK_CLEAR=1;          // unmask ARC0
    }
}    

inline void ARC_end(void)
{
    BVNF_INTR2_2_R5F_MASK_SET=~0;   // mask interrupt 
}   
#endif
